from colouration import Scheme
from .NodeStyle import NodeStyle
from .stylize_edges_based_on_nodes import stylize_edges_based_on_nodes
from .is_root_or_parents_are_in_loop import is_root_or_parents_are_in_loop
from .get_least_used_colour import get_least_used_colour
from .inherit_style import inherit_style
DEFAULT_COLOUR_SCHEME = 'pastel15'


def stylize_with_pensieve(
		graph, node_style=None, edge_style=None, pale_ratio=0.05, divergence_ratio=0.05, edge_darkness_ratio=0.1
):
	if isinstance(graph._colour_scheme, str):
		colour_scheme = Scheme(name=graph._colour_scheme)
	else:
		colour_scheme = graph._colour_scheme
	colours_used = {colour: 0 for colour in colour_scheme.colours}

	# roots or semi-roots
	for node in graph.nodes:
		if is_root_or_parents_are_in_loop(node):
			colour = get_least_used_colour(colours_used)
			if node_style is None:
				style = NodeStyle(colour=colour)
			else:
				style = node_style.copy()
				style.reset_colours()
				style.colour = colour
			node.style = style

	# branch nodes
	for node in graph.nodes:
		if not is_root_or_parents_are_in_loop(node) and not node.is_in_loop():
			node.style = inherit_style(
				node=node, pale_ratio=pale_ratio, divergence_ratio=divergence_ratio, main_style=node_style
			)

	# branch edges
	stylize_edges_based_on_nodes(graph=graph, edge_style=edge_style, edge_darkness_ratio=edge_darkness_ratio)



import random
from colouration import Scheme
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle
from .stylize_edges_based_on_nodes import stylize_edges_based_on_nodes
from .get_least_used_colour import get_least_used_colour
from .inherit_style import inherit_style
DEFAULT_COLOUR_SCHEME = 'pastel15'


def stylize_randomly(
		graph, node_style=None, edge_style=None, pale_ratio=0.05, divergence_ratio=0.05, edge_darkness_ratio=0.1,
		seed=42
):
	if isinstance(graph._colour_scheme, str):
		colour_scheme = Scheme(name=graph._colour_scheme)
	else:
		colour_scheme = graph._colour_scheme
	colours_used = {colour: 0 for colour in colour_scheme.colours}

	nodes = list(graph.nodes)
	random.Random(seed).shuffle(nodes)

	# all nodes
	for node in nodes:
		colour = get_least_used_colour(colours_used)
		if node_style is None:
			style = NodeStyle(colour=colour)
		else:
			style = node_style.copy()
			style.reset_colours()
			style.colour = colour
		node.style = style

	# branch edges
	stylize_edges_based_on_nodes(graph=graph, edge_style=edge_style, edge_darkness_ratio=edge_darkness_ratio)



from .EdgeStyle import EdgeStyle


def stylize_edges_based_on_nodes(graph, edge_style=None, edge_darkness_ratio=0.1):
	for edge in graph.edges:
		colour = edge.start.style.colour.darken(ratio=edge_darkness_ratio)

		if edge_style is None:
			edge_style = EdgeStyle(colour=colour)
		else:
			edge_style = edge_style.copy()
			edge_style.reset_colours()
			edge_style.colour = colour

		if edge.value is not None:
			edge_style._arrow_size *= edge.value

		edge.style = edge_style


import hashlib
import base32hex


def make_hash_sha256(obj):
	hasher = hashlib.sha256()
	hasher.update(repr(obj).encode())
	return base32hex.b32encode(hasher.digest()).replace('=', '-')


class GraphObj:
	def __init__(self, graph, id, value=None, label=None, tooltip=None, style=None, **kwargs):
		self._graph = graph
		self._raw_id = id
		self._value = value
		self._label = label
		self._tooltip = tooltip

		self._frozen = False
		self._parameters = dict(kwargs)

		self._style = style

	def __getstate__(self):
		return {
			'graph': None,
			'id': self.raw_id,
			'value': self._value,
			'label': self._label,
			'tooltip': self._tooltip,
			'style': self._style,
			'frozen': self._frozen,
			'parameters': self._parameters
		}

	def __setstate__(self, state):
		self._graph = state['graph']
		self._raw_id = state['id']
		self._value = state['value']
		self._label = state['label']
		self._tooltip = state['tooltip']
		self._style = state['style']
		self._frozen = state['frozen']
		self._parameters = state['parameters']

	def get(self, item):
		return self._parameters[item]

	def set(self, item, value):
		self._parameters[item] = value

	def __getitem__(self, item):
		return self.get(item=item)

	def __setitem__(self, key, value):
		self.set(item=key, value=value)

	@property
	def graph(self):
		return self._graph

	@property
	def id(self):
		return self.raw_id

	@property
	def raw_id(self):
		return self._raw_id

	@raw_id.setter
	def raw_id(self, raw_id):
		self._raw_id = raw_id

	def __hash__(self):
		return make_hash_sha256((self._graph.__hash__(), self.id)).__hash__()

	@property
	def is_frozen(self):
		return self._frozen

	def freeze(self):
		self._frozen = True

	def unfreeze(self):
		self._frozen = False

	@property
	def value(self):
		return self._value

	@value.setter
	def value(self, value):
		if self.is_frozen:
			raise RuntimeError('GraphObj is frozen!')
		self._value = value

	@property
	def label(self):
		return self._label

	@label.setter
	def label(self, label):
		self._label = label

	@property
	def label_or_value(self):
		if self._label is not None:
			return self._label
		else:
			return self._value

	def __eq__(self, other):
		"""
		:type other: GraphObj
		:rtype: bool
		"""
		if not isinstance(other, self.__class__):
			raise TypeError(f'"{other}" is of type: {type(other)}')
		if self.graph != other.graph:
			raise ValueError('two GraphObj from different graphs cannot be compared')
		return self.id == other.id

	def __ne__(self, other):
		"""
		:type other: GraphObj
		:rtype: bool
		"""
		if not isinstance(other, self.__class__):
			raise TypeError(f'"{other}" is of type: {type(other)}')
		if self.graph != other.graph:
			raise ValueError('two GraphObj from different graphs cannot be compared')
		return self.raw_id != other.raw_id

	def is_similar_to(self, other):
		"""
		:type other: GraphObj
		:rtype: bool
		"""
		condition_1 = self.id == other.id
		condition_2 = self._value == other._value
		condition_3 = self._style == other._style
		condition_4 = self._parameters == other._parameters
		return  condition_1 and condition_2 and condition_3 and condition_4

	def has_style(self):
		return self._style is not None

	@property
	def style(self):
		raise RuntimeError('this parent function is just a placeholder')

	@style.setter
	def style(self, style):
		raise RuntimeError('this parent function is just a placeholder')

	@property
	def graph_style(self):
		# only a place holder, it returns self.graph.edge_style if self is an edge and self.graph.node_style otherwise
		raise RuntimeError('this parent function is just a placeholder')



from ._GraphWithoutDisplay import GraphWithoutDisplay, DEFAULT_PAD


class Graph(GraphWithoutDisplay):
	def get_svg(self, direction=None, pad=DEFAULT_PAD, **kwargs):
		"""
		:type direction: NoneType or str
		:type pad: NoneType or int or float
		:rtype: str
		"""
		return self.render(direction=direction or self._direction, pad=pad, **kwargs)._repr_svg_()

	def _repr_html_(self):
		return self.get_svg()

	def get_html(self, direction=None, pad=DEFAULT_PAD, **kwargs):
		from IPython.core.display import HTML
		return HTML(self.get_svg(direction=direction or self._direction, pad=pad, **kwargs))



from colouration import Colour
from .is_root_or_parents_are_in_loop import is_root_or_parents_are_in_loop


def generate_diverging_numbers(n):
	if n == 1:
		return [0]
	elif n % 2 == 0:
		return [x/(n/2) for x in range(1, n // 2 + 1)] + [-x/(n/2) for x in range(1, n // 2 + 1)]
	else:
		return [0] + generate_diverging_numbers(n-1)


def get_diverging_number(i, n, reverse):
	return sorted(generate_diverging_numbers(n), reverse=reverse)[i]


def inherit_style(node, pale_ratio, divergence_ratio, main_style):
	"""
	:type node: Node
	:rtype: NodeStyle
	"""
	if is_root_or_parents_are_in_loop(node):
		if main_style is None:
			return node.style.copy()
		else:
			return main_style.copy()
	else:
		if len(node.parents) == 1:
			parent = node.parents[0]
			num_siblings = len(parent.children)
			rank = node.get_child_rank(parent=parent)
			if main_style is None:
				style = inherit_style(
					node=parent, pale_ratio=pale_ratio, divergence_ratio=divergence_ratio, main_style=main_style
				)
			else:
				style = main_style
			if num_siblings == 0:
				colour = style.colour.copy(keep_id=True)
			else:
				colour = style.colour.increase_hue(
					amount=divergence_ratio * get_diverging_number(i=rank, n=num_siblings, reverse=False)
				)
			#colour = colour.pale(ratio=pale_ratio)
			colour.saturation = colour.saturation * 3 / 4
			new_style = style.copy()
			new_style.reset_colours()
			new_style.colour = colour
			return new_style
		else:
			styles = [
				inherit_style(node=parent, pale_ratio=pale_ratio, divergence_ratio=0, main_style=main_style)
				for parent in node.parents
				if not node.is_in_loop_with(parent)
			]
			colours = [style.colour for style in styles]
			if main_style is None:
				style = styles[0]
			else:
				style = main_style
			colour = Colour.mix(colours=colours).pale(ratio=pale_ratio)
			new_style = style.copy()
			new_style.reset_colours()
			new_style.colour = colour
			return new_style



import re


def parse_indentations(lines, indent=None):
	"""
	Parse an indented outline into (index, level, content, parent) tuples.
	"""
	regex = re.compile(r'^(?P<indent>(?: {4})*)(?P<name>\S.*)')
	stack = []

	for index, line in enumerate(lines):
		line = line.rstrip('\n')
		if indent is None:
			if line.startswith('\t') or line.startswith(' '):
				indent = re.findall(pattern=r'^[ \t]+', string=line)[0]
		if indent is not None:
			line = line.replace(indent, " " * 4, len(line) - len(line.lstrip(indent)))
		if len(line) > 0:
			match = regex.match(line)
			if not match:
				raise ValueError(
					'Indentation not a multiple of spaces or tabs'.format(line)
				)
			level = len(match.group('indent')) // 4
			if level > len(stack):
				raise ValueError('Indentation too deep: "{0}"'.format(line))
			# stack[level:] = [match.group('name')]
			stack[level:] = [index]
			content = match.group('name')
		else:
			content = ''
		yield index, level, content, (stack[level - 1] if level else None)
